@@ -1,12 +1,13 @@
 #include "config.h"
 #include "aworkspaces.h"
 #include "workspaces.h"
-#include "prefs.h"
+#include "default.h"
+#include "yprefs.h"
 #include "yxapp.h"
 #include "wmframe.h"
 #include "wmmgr.h"
 #include "udir.h"
-#include "ypaths.h"
+#include "wmapp.h"
 #include "wpixmaps.h"
 #include "intl.h"
 #include <math.h>
@@ -19,19 +20,20 @@ YColorName WorkspaceButton::activeButtonBg(&clrWorkspaceActiveButton);
 YColorName WorkspaceButton::activeBackupBg(&clrActiveButton);
 YColorName WorkspaceButton::activeButtonFg(&clrWorkspaceActiveButtonText);
 
-ref<YFont> WorkspaceButton::normalButtonFont;
-ref<YFont> WorkspaceButton::activeButtonFont;
+YFont WorkspaceButton::normalButtonFont;
+YFont WorkspaceButton::activeButtonFont;
 
 WorkspaceButton::WorkspaceButton(int ws, YWindow *parent, WorkspaceDragger* d):
     super(parent, YAction()),
     fWorkspace(ws),
     fDelta(0),
     fDownX(0),
+    fStale(true),
     fDragging(false),
     fGraphics(this, true),
     fPane(d)
 {
-    addStyle(wsNoExpose);
+    addStyle(wsNoExpose | wsToolTipping);
     setParentRelative();
     setTitle(name());
 }
@@ -104,6 +106,8 @@ void WorkspaceButton::handleCrossing(const XCrossingEvent &e) {
 
     if (false == pagerShowPreview) {
         super::handleCrossing(e);
+    } else {
+        YWindow::handleCrossing(e);
     }
 }
 
@@ -118,14 +122,16 @@ void WorkspaceButton::handleDNDLeave() {
     repaint();
 }
 
-void WorkspaceButton::handleBeginDrag(const XButtonEvent& d, const XMotionEvent& m)
+bool WorkspaceButton::handleBeginDrag(const XButtonEvent& d, const XMotionEvent& m)
 {
-    if (d.button == Button1) {
+    if (d.button == Button1 && fPane->limited()) {
         fDragging = true;
         fDownX = d.x_root;
         fDelta = m.x_root - d.x_root;
         fPane->drag(fWorkspace, fDelta, true, false);
+        return true;
     }
+    return false;
 }
 
 void WorkspaceButton::handleDrag(const XButtonEvent& d, const XMotionEvent& m)
@@ -162,6 +168,7 @@ void WorkspaceButton::inputReturn(YInputLine* input) {
             swap(*&str, *workspaces[fWorkspace]);
             manager->setDesktopNames();
             fPane->relabel(fWorkspace);
+            repaint();
         }
         fInput = nullptr;
     }
@@ -212,10 +219,25 @@ WorkspacesPane::WorkspacesPane(YWindow *parent):
     fTime(zerotime()),
     fMillis(16L),
     fRepositioning(false),
-    fReconfiguring(false)
+    fReconfiguring(false),
+    fRepaintSpaces(false),
+    fButtons(workspaceCount)
 {
     addStyle(wsNoExpose);
     setParentRelative();
+    createButtons();
+}
+
+void WorkspacesPane::createButtons() {
+    fReconfiguring = true;
+    unsigned width = 0;
+    unsigned height = smallIconSize + 8;
+    for (int i = 0, n = workspaceCount; i < n; ++i) {
+        width += create(i, height)->width();
+    }
+    resize(width, height);
+    setPressed(manager->activeWorkspace(), true);
+    fReconfiguring = false;
 }
 
 void WorkspacesPane::resize(unsigned width, unsigned height) {
@@ -236,6 +258,8 @@ long WorkspacesPane::limitWidth(long paneWidth) {
     const char* str = taskBarWorkspacesLimit;
     long taskBarWidth = desktop->getScreenGeometry().width()
                       * taskBarWidthPercentage / 100;
+    long horizontally = taskBarWorkspacesLeft == leftToRight ?
+                        taskBarWidth - x() : x() + long(width());
     long reserved = 200L;
     long maxPixels = 0;
     long maxButtons = 0;
@@ -244,7 +268,7 @@ long WorkspacesPane::limitWidth(long paneWidth) {
         char* end = nullptr;
         long num = strtol(str, &end, 0);
         if (end && str < end && inrange(num, 0L, long(SHRT_MAX))) {
-            maxPixels = max(50L, taskBarWidth - x() - reserved);
+            maxPixels = max(50L, horizontally - reserved);
             maxButtons = maxPixels * count() / non_zero(paneWidth);
             maxPercent = 100L * maxPixels / taskBarWidth;
             if (*end == ' ')
@@ -264,9 +288,9 @@ long WorkspacesPane::limitWidth(long paneWidth) {
     }
     if (isEmpty(str) || 0 == (maxPixels | maxButtons)) {
         if (taskBarDoubleHeight && taskBarWorkspacesTop) {
-            maxPixels = taskBarWidth - x() - reserved;
+            maxPixels = horizontally - reserved;
         } else {
-            maxPixels = (taskBarWidth - x() - reserved) / 2;
+            maxPixels = (horizontally - reserved) / 2;
             maxButtons = 20;    // old limit
         }
     }
@@ -342,19 +366,7 @@ void WorkspacesPane::relabelButtons() {
 
 void WorkspacesPane::configure(const YRect2& r) {
     if ((fReconfiguring | fRepositioning) == false && r.resized()) {
-        fReconfiguring = true;
-        if (count() == 0) {
-            unsigned width = 0, height = max(smallIconSize + 8, r.height());
-            for (int i = 0, n = workspaceCount; i < n; ++i)
-                width += create(i, height)->width();
-            resize(width, height);
-            setPressed(manager->activeWorkspace(), true);
-            paths = null;
-        }
-        else {
-            repositionButtons();
-        }
-        fReconfiguring = false;
+        repositionButtons();
     }
 }
 
@@ -362,8 +374,9 @@ void WorkspacesPane::updateButtons() {
     MSG(("WorkspacesPane::udpateButtons(): updating %d -> %d",
          count(), int(workspaceCount)));
 
-    if (count() > workspaceCount)
-        fButtons.shrink(max<int>(1, workspaceCount));
+    while (workspaceCount < count() && 1 < count()) {
+        fButtons.remove(rightToLeft ? 0 : count() - 1);
+    }
 
     int width = extent() - fMoved;
     for (int i = count(), n = workspaceCount; i < n; ++i) {
@@ -385,7 +398,7 @@ void WorkspacesPane::updateButtons() {
 
 WorkspaceButton* WorkspacesPane::create(int workspace, unsigned height) {
     WorkspaceButton *wk = new WorkspaceButton(workspace, this, this);
-    fButtons += wk;
+    fButtons.insert(rightToLeft ? 0 : count(), wk);
     if (pagerShowPreview) {
         unsigned dw = desktop->width();
         unsigned dh = desktop->height();
@@ -401,9 +414,10 @@ WorkspaceButton* WorkspacesPane::create(int workspace, unsigned height) {
 }
 
 WorkspaceIcons::WorkspaceIcons() {
-    ref<YResourcePaths> dirs(YResourcePaths::subdirs("workspace", false));
-    for (int i = 0; i < dirs->getCount(); ++i) {
-        upath path(dirs->getPath(i) + "/workspace/");
+    MStringArray dirs;
+    wmapp->subdirs("workspace", false, dirs);
+    for (const mstring& base : dirs) {
+        upath path(base + "/workspace/");
         for (adir dir(path.string()); dir.next(); ) {
             files += (path + dir.entry()).string();
         }
@@ -437,17 +451,17 @@ void WorkspacesPane::label(WorkspaceButton* wk) {
 }
 
 void WorkspacesPane::relabel(int ws) {
-    label(fButtons[ws]);
+    label(index(ws));
     repositionButtons();
     paths = null;
 }
 
-void WorkspacesPane::setPressed(long ws, bool set) {
-    if (inrange<long>(1+ws, 1, count())) {
-        WorkspaceButton* wk = fButtons[int(ws)];
+void WorkspacesPane::setPressed(int ws, bool set) {
+    if (inrange(ws, 0, count() - 1)) {
+        WorkspaceButton* wk = index(ws);
         wk->setPressed(set);
         if (set) {
-            fActive = int(ws);
+            fActive = ws;
             if (wk->extent() > int(width())) {
                 fMoved -= wk->extent() - int(width());
                 repositionButtons();
@@ -460,6 +474,10 @@ void WorkspacesPane::setPressed(long ws, bool set) {
     }
 }
 
+bool WorkspacesPane::limited() const {
+    return 0 < count() && (last()->extent() - fButtons[0]->x()) > int(width());
+}
+
 void WorkspacesPane::drag(int ws, int dx, bool start, bool end) {
     int previous = fMoved;
     int farthest = extent() - fButtons[0]->x();
@@ -507,8 +525,9 @@ bool WorkspacesPane::handleTimer(YTimer *t) {
         if (fRepaintSpaces) {
             fRepaintSpaces = false;
             if (pagerShowPreview) {
-                for (IterType wk = iterator(); ++wk; ) {
-                    if (wk->visible() &&
+                for (auto wk : fButtons) {
+                    if (wk->isStale() &&
+                        wk->visible() &&
                         wk->extent() > 0 &&
                         wk->x() < int(width()))
                     {
@@ -528,10 +547,10 @@ void WorkspacesPane::stopDrag() {
     }
 }
 
-ref<YFont> WorkspaceButton::getActiveFont() {
+YFont WorkspaceButton::getActiveFont() {
     if (activeButtonFont == null) {
-        if (*activeWorkspaceFontName || *activeWorkspaceFontNameXft) {
-            activeButtonFont = YFont::getFont(XFA(activeWorkspaceFontName));
+        if (activeWorkspaceFontName.nonempty()) {
+            activeButtonFont = activeWorkspaceFontName;
         }
         if (activeButtonFont == null) {
             activeButtonFont = YButton::getActiveFont();
@@ -540,13 +559,13 @@ ref<YFont> WorkspaceButton::getActiveFont() {
     return activeButtonFont;
 }
 
-ref<YFont> WorkspaceButton::getFont() {
+YFont WorkspaceButton::getFont() {
     if (isPressed()) {
         return getActiveFont();
     }
     else if (normalButtonFont == null) {
-        if (*normalWorkspaceFontName || *normalWorkspaceFontNameXft) {
-            normalButtonFont = YFont::getFont(XFA(normalWorkspaceFontName));
+        if (normalWorkspaceFontName.nonempty()) {
+            normalButtonFont = normalWorkspaceFontName;
         }
         if (normalButtonFont == null) {
             normalButtonFont = YButton::getFont();
@@ -572,7 +591,7 @@ YSurface WorkspaceButton::getSurface() {
 }
 
 YDimension WorkspaceButton::getTextSize() {
-    ref<YFont> font(getActiveFont());
+    YFont font(getActiveFont());
     return YDimension(font->textWidth(name()), font->height());
 }
 
@@ -589,18 +608,37 @@ mstring WorkspaceButton::baseName() {
     return name;
 }
 
-void WorkspaceButton::updateName() {
+void WorkspaceButton::updateToolTip() {
     setToolTip(_("Workspace: ") + baseName());
 }
 
-void WorkspacesPane::repaint() {
-    if (!pagerShowPreview) return;
+void WorkspaceButton::updateName() {
+    if (toolTipVisible()) {
+        updateToolTip();
+    }
+}
 
-    fRepaintSpaces = true;
-    fRepaintTimer->setTimer(20L, this, true);
+void WorkspacesPane::repaint() {
+    if (pagerShowPreview) {
+        fRepaintSpaces = true;
+        fRepaintTimer->setTimer(20L, this, true);
+    }
+}
+
+void WorkspacesPane::repaintWorkspace(int ws) {
+    if (pagerShowPreview) {
+        if (ws == AllWorkspaces) {
+            ws = manager->activeWorkspace();
+        }
+        if (inrange(ws, 0, count() - 1)) {
+            fButtons[ws]->setStale();
+        }
+    }
 }
 
 void WorkspaceButton::paint(Graphics &g, const YRect& r) {
+    fStale = false;
+
     if (!pagerShowPreview) {
         YButton::paint(g, r);
         return;
@@ -696,16 +734,17 @@ void WorkspaceButton::paint(Graphics &g, const YRect& r) {
             strlcpy(label, baseName(), min(5, int(sizeof label)));
         }
         if (label[0] != 0) {
-            ref<YFont> font = getFont();
+            YFont font = getFont();
+            if (font) {
+                int wx = (w - font->textWidth(label)) / 2 + x;
+                int wy = (h - font->height()) / 2 + font->ascent() + y;
 
-            int wx = (w - font->textWidth(label)) / 2 + x;
-            int wy = (h - font->height()) / 2 + font->ascent() + y;
-
-            g.setFont(font);
-            g.setColor(colors[0]);
-            g.drawChars(label, 0, strlen(label), wx+1, wy+1);
-            g.setColor(colors[3]);
-            g.drawChars(label, 0, strlen(label), wx, wy);
+                g.setFont(font);
+                g.setColor(colors[0]);
+                g.drawString(wx+1, wy+1, label);
+                g.setColor(colors[3]);
+                g.drawString(wx, wy, label);
+            }
         }
         if (surface.gradient != null) {
             g.maxOpacity();
