@@ -20,8 +20,10 @@
 #include "sysdep.h"
 #include "intl.h"
 #include "appnames.h"
+#include <ctype.h>
+#include "debug.h"
 
-char const *ApplicationName(0);
+char const *ApplicationName(nullptr);
 
 #ifndef LPCSTR // mind the MFC
 // easier to read...
@@ -36,7 +38,12 @@ typedef const char* LPCSTR;
 #include "ycollections.h"
 
 // program options
-bool add_sep_before(false), add_sep_after(false), no_sep_others(false), no_sub_cats(false);
+bool add_sep_before = false;
+bool add_sep_after = false;
+bool no_sep_others = false;
+bool no_sub_cats = false;
+bool generic_name = false;
+bool right_to_left = false;
 
 template<typename T, void TFreeFunc(T)>
 struct auto_raii {
@@ -59,35 +66,32 @@ static int cmpUtf8(const void *p1, const void *p2) {
 class tDesktopInfo;
 typedef YVec<const gchar*> tCharVec;
 tCharVec sys_folders, home_folders;
-tCharVec* sys_home_folders[] = { &sys_folders, &home_folders, 0 };
-tCharVec* home_sys_folders[] = { &home_folders, &sys_folders, 0 };
+tCharVec* sys_home_folders[] = { &sys_folders, &home_folders, nullptr };
+tCharVec* home_sys_folders[] = { &home_folders, &sys_folders, nullptr };
 
 struct tListMeta {
     LPCSTR title, key, icon;
     LPCSTR const * const parent_sec;
-#if 0
-    static int cmpTitleUtf8(const void *a, const void *b) {
-        tListMeta *A((tListMeta*) a), *B((tListMeta*) b);
-        return cmpUtf8(A->title, B->title);
-    }
-
-    static int cmpTitle(const void *a, const void *b) {
-        tListMeta *A((tListMeta*) a), *B((tListMeta*) b);
-        return strcmp(A->title, B->title);
-    }
-
-    static int cmpCategoryThroughPtr(const void *a, const void *b) {
-        tListMeta **A((tListMeta**) a), **B((tListMeta**) b);
-        return strcmp((**A).key, (**B).key);
-    }
-#endif
+    enum eLoadFlags {
+        NONE = 0,
+        DEFAULT_TRANSLATED = 1,
+        SYSTEM_TRANSLATED = 2,
+        DEFAULT_ICON = 4,
+        FALLBACK_ICON = 8,
+        SYSTEM_ICON = 16,
+    };
+    short load_state_icon, load_state_title;
 };
 GHashTable* meta_lookup_data;
 tListMeta* lookup_category(LPCSTR key)
 {
     tListMeta* ret = (tListMeta*) g_hash_table_lookup(meta_lookup_data, key);
-    if(ret && ret->title == NULL)
-        ret->title = _(ret->key);
+    if (ret)
+    {
+        // auto-translate the default title for the user's language
+        if (ret->title == nullptr)
+            ret->title = _(ret->key);
+    }
     return ret;
 }
 
@@ -101,10 +105,12 @@ struct t_menu_node {
 protected:
     // for leafs -> NULL, otherwise sub-menu contents
     GTree* store;
-    LPCSTR progCmd;
+    char* progCmd;
+    const char* generic;
 public:
     const tListMeta *meta;
-    t_menu_node(const tListMeta* desc): store(0), progCmd(0), meta(desc) {}
+    t_menu_node(const tListMeta* desc):
+        store(nullptr), progCmd(nullptr), generic(nullptr), meta(desc) {}
 
     struct t_print_meta {
         int count, level;
@@ -116,18 +122,25 @@ public:
     }
 
     void print(t_print_meta *ctx) {
-        if(!meta) return;
+        if (!meta) return;
         LPCSTR title = Elvis(meta->title, meta->key);
 
-        if(!store)
+        if (!store)
         {
-            if(title && progCmd) {
-                if(ctx->count == 0 && add_sep_before)
+            if (title && progCmd) {
+                if (ctx->count == 0 && add_sep_before)
                     puts("separator");
-                printf("prog \"%s\" %s %s\n",
-                        title,
-                        meta->icon,
-                        progCmd);
+                if (nonempty(generic) && !strcasestr(title, generic)) {
+                    if (right_to_left) {
+                        printf("prog \"(%s) %s\" %s %s\n",
+                                generic, title, meta->icon, progCmd);
+                    } else {
+                        printf("prog \"%s (%s)\" %s %s\n",
+                                title, generic, meta->icon, progCmd);
+                    }
+                }
+                else
+                    printf("prog \"%s\" %s %s\n", title, meta->icon, progCmd);
             }
             ctx->count++;
             return;
@@ -151,7 +164,7 @@ public:
         }
         ctx->level++;
         g_tree_foreach(store, print_node, ctx);
-        if(ctx->level == 1 && ctx->print_separated)
+        if (ctx->level == 1 && ctx->print_separated)
         {
             puts("separator");
             no_sep_others = true;
@@ -164,7 +177,7 @@ public:
 #else
             printf("# end of menu \"%s\"\n}\n", title);
 #endif
-        if(add_sep_after && ctx->level == 0 && ctx->count > 0)
+        if (add_sep_after && ctx->level == 0 && ctx->count > 0)
             puts("separator");
 
     }
@@ -173,14 +186,15 @@ public:
      * Usual print method for the root node
      */
     void print() {
-        t_print_meta ctx = {0,0,0};
+        t_print_meta ctx = {0,0,nullptr};
         print(&ctx);
     }
 
     void add(t_menu_node* node) {
         if (!store)
             store = g_tree_new(cmpUtf8);
-        g_tree_replace(store, (gpointer) Elvis(node->meta->title, node->meta->key), (gpointer) node);
+        if (node->meta->title || node->meta->key)
+            g_tree_replace(store, (gpointer) Elvis(node->meta->title, node->meta->key), (gpointer) node);
     }
 
     /**
@@ -209,8 +223,8 @@ public:
         // skip the rest of the further nesting (cannot fit into any)
         bool skipping = false;
 
-        tListMeta* pNewCatInfo = 0;
-        tListMeta** ppLastMainCat = 0;
+        tListMeta* pNewCatInfo = nullptr;
+        tListMeta** ppLastMainCat = nullptr;
 
         for (const char * const *pSubCatName = subCatCandidate->parent_sec;
                 *pSubCatName; ++pSubCatName) {
@@ -225,7 +239,7 @@ public:
                 pTree->get_subtree(subCatCandidate)->add(pNode);
                 // main menu was served, don't come here again
                 if (ppLastMainCat)
-                    *ppLastMainCat = 0;
+                    *ppLastMainCat = nullptr;
             }
 
             skipping = true;
@@ -252,9 +266,9 @@ public:
             if (skipping)
                 continue;
             // if not on first node, make or find submenues for the comming tokens
-            if(!pNewCatInfo)
+            if (!pNewCatInfo)
                 pNewCatInfo = lookup_category(*pSubCatName);
-            if(!pNewCatInfo)
+            if (!pNewCatInfo)
                 return; // heh? fantasy category? Let caller handle it
             pTree = pTree->get_subtree(pNewCatInfo);
         }
@@ -268,14 +282,14 @@ public:
                 continue; // empty?
             tListMeta *pResolved = lookup_category(*pCatKey);
             if (!pResolved) continue;
-            if(!pResolved->parent_sec)
+            if (!pResolved->parent_sec)
                 matched_main_cats.add(pResolved);
             else
                 matched_sub_cats.add(pResolved);
         }
         if (matched_main_cats.size == 0)
             matched_main_cats.add(lookup_category("Other"));
-        if(!no_sub_cats) {
+        if (!no_sub_cats) {
             for (tListMeta** p = matched_sub_cats.data;
                     p < matched_sub_cats.data + matched_sub_cats.size; ++p) {
 
@@ -285,7 +299,7 @@ public:
         for (tListMeta** p = matched_main_cats.data;
                 p < matched_main_cats.data + matched_main_cats.size; ++p) {
 
-            if (*p == NULL)
+            if (*p == nullptr)
                 continue;
             get_subtree(*p)->add(pNode);
         }
@@ -319,7 +333,7 @@ public:
 // tear down if not permitted
         if (!g_app_info_should_show(*this)) {
             g_object_unref(pInfo);
-            pInfo = 0;
+            pInfo = nullptr;
             return;
         }
     }
@@ -336,10 +350,16 @@ public:
 
     LPCSTR get_name() const {
         if (!pInfo)
-            return 0;
+            return nullptr;
         return g_app_info_get_display_name((GAppInfo*) pInfo);
     }
 
+    LPCSTR get_generic() const {
+        if (!pInfo || !generic_name)
+            return nullptr;
+        return g_desktop_app_info_get_generic_name(pInfo);
+    }
+
     char * get_icon_path() const {
         GIcon *pIcon = g_app_info_get_icon((GAppInfo*) pInfo);
         auto_gunref free_icon((GObject*) pIcon);
@@ -347,18 +367,18 @@ public:
         if (pIcon) {
             char *icon_path = g_icon_to_string(pIcon);
             if (!icon_path)
-                return 0;
+                return nullptr;
             // if absolute then we are done here
             if (icon_path[0] == '/')
                 return icon_path;
             // err, not owned! auto_gfree free_orig_icon_path(icon_path);
             return icon_path;
         }
-        return 0;
+        return nullptr;
     }
 };
 
-tListMeta no_description = {0,0,0,0};
+tListMeta no_description = {nullptr,nullptr,nullptr,nullptr};
 
 t_menu_node root(&no_description);
 
@@ -432,27 +452,100 @@ struct t_menu_node_app : t_menu_node
         else
             // not simple command or needs a terminal started via launcher callback, or both
             progCmd = g_strdup_printf("%s \"%s\"", ApplicationName, dinfo.d_file);
+        if (cmdMod && cmdMod != progCmd)
+            g_free(cmdMod);
+        generic = dinfo.get_generic();
+    }
+    ~t_menu_node_app() {
+        g_free(progCmd);
     }
 };
 
-typedef void (*tFuncInsertInfo)(const char* szDesktopFile);
-void pickup_folder_info(const char* szDesktopFile) {
+struct tFromTo { LPCSTR from; LPCSTR to;};
+// match transformations applied by some DEs
+const tFromTo SameCatMap[] = {
+        {"Utilities", "Utility"},
+        {"Sound & Video", "AudioVideo"},
+        {"Sound", "Audio"},
+        {"File", "FileTools"},
+        {"Terminal Applications", "TerminalEmulator"},
+        {"Mathematics", "Math"},
+        {"Arcade", "ArcadeGame"},
+        {"Card Games", "CardGame"}
+};
+const tFromTo SameIconMap[] = {
+        { "AudioVideo", "Audio" },
+        { "AudioVideo", "Video" }
+};
+typedef void (*tFuncInsertInfo)(LPCSTR szDesktopFile);
+void pickup_folder_info(LPCSTR szDesktopFile) {
     GKeyFile *kf = g_key_file_new();
     auto_raii<GKeyFile*, g_key_file_free> free_kf(kf);
 
-    if (!g_key_file_load_from_file(kf, szDesktopFile, G_KEY_FILE_NONE, 0))
+    if (!g_key_file_load_from_file(kf, szDesktopFile, G_KEY_FILE_NONE, nullptr))
         return;
-    const char* icon_name = g_key_file_get_string(kf, "Desktop Entry", "Icon",
-            0);
-    const char* cat_name = g_key_file_get_string(kf, "Desktop Entry", "Name",
-            0);
-    const char* cat_title = g_key_file_get_locale_string(kf, "Desktop Entry",
-            "Name", NULL, NULL);
+    LPCSTR cat_name = g_key_file_get_string(kf, "Desktop Entry", "Name", nullptr);
+    // looks like bad data
+    if (!cat_name || !*cat_name)
+        return;
+    // try a perfect match by name or file name
     tListMeta* pCat = lookup_category(cat_name);
+    if (!pCat)
+    {
+        gchar* bn(g_path_get_basename(szDesktopFile));
+        auto_gfree cleanr(bn);
+        char* dot = strchr(bn, '.');
+        if (dot)
+        {
+            *dot = 0x0;
+            pCat = lookup_category(bn);
+        }
+    }
+    for (const tFromTo* p = SameCatMap;
+            !pCat && p < SameCatMap+ACOUNT(SameCatMap);
+            ++p) {
+
+        if (0 == strcmp(cat_name, p->from))
+            pCat = lookup_category(p->to);
+    }
+
     if (!pCat)
         return;
-    pCat->icon = icon_name;
-    pCat->title = cat_title;
+    if (pCat->load_state_icon < tListMeta::SYSTEM_ICON) {
+        LPCSTR icon_name = g_key_file_get_string (kf, "Desktop Entry",
+                                                       "Icon", nullptr);
+        if (icon_name && *icon_name) {
+            pCat->icon = icon_name;
+            pCat->load_state_icon = tListMeta::SYSTEM_ICON;
+        }
+    }
+    if (pCat->load_state_title < tListMeta::SYSTEM_TRANSLATED) {
+        char* cat_title = g_key_file_get_locale_string(kf, "Desktop Entry",
+                                                       "Name", nullptr, nullptr);
+        if (!cat_title) return;
+        pCat->title = cat_title;
+        char* cat_title_c = g_key_file_get_string(kf, "Desktop Entry",
+                                                  "Name", nullptr);
+        bool same_trans = 0 == strcmp (cat_title_c, cat_title);
+        if (!same_trans) pCat->load_state_title = tListMeta::SYSTEM_TRANSLATED;
+        // otherwise: not sure, keep searching for a better translation
+    }
+    // something special, donate the icon to similar items unless they have a better one
+
+    for (const tFromTo* p = SameIconMap; p < SameIconMap + ACOUNT(SameIconMap);
+            ++p) {
+
+        if (strcmp (pCat->key, p->from))
+            continue;
+
+        tListMeta *t = lookup_category (p->to);
+        // give them at least some icon for now
+        if (t && t->load_state_icon < tListMeta::FALLBACK_ICON) {
+            t->icon = pCat->icon;
+            t->load_state_icon = tListMeta::FALLBACK_ICON;
+        }
+
+    }
 }
 
 void insert_app_info(const char* szDesktopFile) {
@@ -480,21 +573,21 @@ void proc_dir_rec(LPCSTR syspath, unsigned depth,
         LPCSTR szFileSfx) {
     gchar *path = g_strjoin("/", syspath, szSubfolder, NULL);
     auto_gfree relmem_path(path);
-    GDir *pdir = g_dir_open(path, 0, NULL);
+    GDir *pdir = g_dir_open(path, 0, nullptr);
     if (!pdir)
         return;
     auto_raii<GDir*, g_dir_close> dircloser(pdir);
 
-    const gchar *szFilename(NULL);
-    while (NULL != (szFilename = g_dir_read_name(pdir))) {
+    const gchar *szFilename(nullptr);
+    while (nullptr != (szFilename = g_dir_read_name(pdir))) {
         if (!szFilename || !checkSuffix(szFilename, szFileSfx))
             continue;
 
         gchar *szFullName = g_strjoin("/", path, szFilename, NULL);
         auto_gfree xxfree(szFullName);
         static GStatBuf buf;
-        if (g_stat(szFullName, &buf))
-            return;
+        if (0 != g_stat(szFullName, &buf))
+            continue;
         if (S_ISDIR(buf.st_mode)) {
             static ino_t reclog[6];
             for (unsigned i = 0; i < depth; ++i) {
@@ -517,16 +610,16 @@ void proc_dir_rec(LPCSTR syspath, unsigned depth,
     }
 }
 
-bool launch(LPCSTR dfile, LPCSTR *argv, int argc) {
+bool launch(LPCSTR dfile, char** argv, int argc) {
     GDesktopAppInfo *pInfo = g_desktop_app_info_new_from_filename(dfile);
     if (!pInfo)
         return false;
 #if 0 // g_file_get_uri crashes, no idea why, even enforcing file prefix doesn't help
-    if (argc>0)
+    if (argc > 0)
     {
-        GList* parms=NULL;
-        for (int i=0; i<argc; ++i)
-        parms=g_list_append(parms,
+        GList* parms = NULL;
+        for (int i = 0; i < argc; ++i)
+        parms = g_list_append(parms,
                 g_strdup_printf("%s%s", strstr(argv[i], "://") ? "" : "file://",
                         argv[i]));
         return g_app_info_launch ((GAppInfo *)pInfo,
@@ -537,14 +630,31 @@ bool launch(LPCSTR dfile, LPCSTR *argv, int argc) {
     (void) argv;
     (void) argc;
 #endif
-    return g_app_info_launch((GAppInfo *) pInfo,
-    NULL,
-    NULL, NULL);
+    return g_app_info_launch((GAppInfo *) pInfo, nullptr, nullptr, nullptr);
 }
 
 static void init() {
 #ifdef CONFIG_I18N
-    setlocale(LC_ALL, "");
+    const char* loc = setlocale(LC_ALL, "");
+    if (loc
+        && islower(*loc & 0xff)
+        && islower(loc[1] & 0xff)
+        && !isalpha(loc[2] & 0xff)) {
+        const char rtls[][4] = {
+            "ar",   // arabic
+            "fa",   // farsi
+            "he",   // hebrew
+            "ps",   // pashto
+            "sd",   // sindhi
+            "ur",   // urdu
+        };
+        for (const char* rtl : rtls) {
+            if (rtl[0] == loc[0] && rtl[1] == loc[1]) {
+                right_to_left = true;
+                break;
+            }
+        }
+    }
 #endif
 
     bindtextdomain(PACKAGE, LOCDIR);
@@ -554,7 +664,7 @@ static void init() {
 
     for (unsigned i = 0; i < ACOUNT(spec::menuinfo); ++i) {
         tListMeta& what = spec::menuinfo[i];
-        if(no_sub_cats && what.parent_sec)
+        if (no_sub_cats && what.parent_sec)
             continue;
         // enforce non-const since we are not destroying that data ever, no key_destroy_func set!
         g_hash_table_insert(meta_lookup_data, (gpointer) what.key, &what);
@@ -563,17 +673,20 @@ static void init() {
 
 static void help(LPCSTR home, LPCSTR dirs, FILE* out, int xit) {
     g_fprintf(out,
-            "USAGE: icewm-menu-fdo OPTIONS\n"
+            "USAGE: icewm-menu-fdo [OPTIONS] [FILENAME]\n"
             "OPTIONS:\n"
+            "-g, --generic\tInclude GenericName in parentheses of progs\n"
+            "-o, --output=FILE\tWrite the output to FILE\n"
             "--seps  \tPrint separators before and after contents\n"
+            "--sep-before\tPrint separator before the contents\n"
             "--sep-after\tPrint separator only after contents\n"
             "--no-sep-others\tNo separation of the 'Others' menu point\n"
             "--no-sub-cats\tNo additional subcategories, just one level of menues\n"
-            "*.desktop\tAny .desktop file to launch the application command from there\n"
-            "This program also listens to "
-                    "environment variables defined by the\nXDG Base Directory Specification:\n"
-                    "XDG_DATA_HOME=%s\n"
-                    "XDG_DATA_DIRS=%s\n", home, dirs);
+            "FILENAME\tAny .desktop file to launch its application Exec command\n"
+            "This program also listens to environment variables defined by\n"
+            "the XDG Base Directory Specification:\n"
+            "XDG_DATA_HOME=%s\n"
+            "XDG_DATA_DIRS=%s\n", home, dirs);
     exit(xit);
 }
 
@@ -601,16 +714,30 @@ void load_folder_descriptions(const tCharVec& where) {
     }
 }
 
-int main(int argc, LPCSTR *argv) {
+#ifdef DEBUG_xxx
+void dbgPrint(const gchar *msg)
+{
+    tlog("%s", msg);
+}
+#endif
+
+int main(int argc, char** argv) {
     ApplicationName = my_basename(argv[0]);
 
 #if !GLIB_CHECK_VERSION(2,36,0)
     g_type_init();
 #endif
 
-    LPCSTR usershare = getenv("XDG_DATA_HOME");
+#ifdef DEBUG_xxx
+    // XXX: if enabled, one probably also need to enable debug facilities in its code, like
+    // what --debug option does
+    g_set_printerr_handler(dbgPrint);
+    g_set_print_handler(dbgPrint);
+#endif
+
+    char* usershare = getenv("XDG_DATA_HOME");
     if (!usershare || !*usershare)
-        usershare = g_strjoin(NULL, getenv("HOME"), "/.local/share", NULL);
+        usershare = g_strjoin(nullptr, getenv("HOME"), "/.local/share", NULL);
 
     // system dirs, either from environment or from static locations
     LPCSTR sysshare = getenv("XDG_DATA_DIRS");
@@ -622,33 +749,43 @@ int main(int argc, LPCSTR *argv) {
         return EXIT_SUCCESS;
     }
 
-    for (LPCSTR *pArg = argv + 1; pArg < argv + argc; ++pArg) {
+    for (char** pArg = argv + 1; pArg < argv + argc; ++pArg) {
         if (is_version_switch(*pArg))
             print_version_exit(VERSION);
-        if (is_help_switch(*pArg))
+        else if (is_copying_switch(*pArg))
+            print_copying_exit();
+        else if (is_help_switch(*pArg))
             help(usershare, sysshare, stdout, EXIT_SUCCESS);
-        if (is_long_switch(*pArg, "seps")) {
+        else if (is_long_switch(*pArg, "seps"))
             add_sep_before = add_sep_after = true;
-            continue;
-        }
-        if (is_long_switch(*pArg, "sep-before")) {
+        else if (is_long_switch(*pArg, "sep-before"))
             add_sep_before = true;
-            continue;
-        }
-        if (is_long_switch(*pArg, "sep-after")) {
+        else if (is_long_switch(*pArg, "sep-after"))
             add_sep_after = true;
-            continue;
-        }
-        if (is_long_switch(*pArg, "no-sep-others")) {
+        else if (is_long_switch(*pArg, "no-sep-others"))
             no_sep_others = true;
-            continue;
-        }
-        if (is_long_switch(*pArg, "no-sub-cats")) {
+        else if (is_long_switch(*pArg, "no-sub-cats"))
             no_sub_cats = true;
-            continue;
+        else if (is_switch(*pArg, "g", "generic-name"))
+            generic_name = true;
+        else {
+            char *value = nullptr, *expand = nullptr;
+            if (GetArgument(value, "o", "output", pArg, argv + argc)) {
+                if (*value == '~')
+                    value = expand = tilde_expansion(value);
+                else if (*value == '$')
+                    value = expand = dollar_expansion(value);
+                if (nonempty(value)) {
+                    if (freopen(value, "w", stdout) == nullptr) {
+                        fflush(stdout);
+                    }
+                }
+                if (expand)
+                    delete[] expand;
+            }
+            else // unknown option
+                help(usershare, sysshare, stderr, EXIT_FAILURE);
         }
-        // unknown option?
-        help(usershare, sysshare, stderr, EXIT_FAILURE);
     }
 
     init();
@@ -663,6 +800,9 @@ int main(int argc, LPCSTR *argv) {
 
     root.print();
 
+    if (nonempty(usershare) && usershare != getenv("XDG_DATA_HOME"))
+        g_free(usershare);
+
     return EXIT_SUCCESS;
 }
 
