@@ -74,34 +74,36 @@ static bool post_preferences;
 static bool show_extensions;
 
 static Window registerProtocols1(char **argv, int argc) {
-    long timestamp = CurrentTime;
+    const long timestamp = CurrentTime;
+    const Window xroot = xapp->root();
     YAtom wmSx("WM_S", true);
-
-    Window current_wm = XGetSelectionOwner(xapp->display(), wmSx);
-
-    if (current_wm != None) {
-        if (!replace_wm)
-            die(1, _("A window manager is already running, use --replace to replace it"));
-      XSetWindowAttributes attrs;
-      attrs.event_mask = StructureNotifyMask;
-      XChangeWindowAttributes (
-          xapp->display(), current_wm,
-          CWEventMask, &attrs);
+    const Window current_wm = XGetSelectionOwner(xapp->display(), wmSx);
+    if (current_wm) {
+        if (replace_wm) {
+            XSetWindowAttributes attr;
+            attr.event_mask = StructureNotifyMask;
+            XChangeWindowAttributes(xapp->display(), current_wm,
+                                    CWEventMask, &attr);
+        } else {
+            die(1, _("A window manager is already running, "
+                     "use --replace to replace it"));
+        }
     }
 
-    Window xroot = xapp->root();
-    Window xid =
-        XCreateSimpleWindow(xapp->display(), xroot,
-            0, 0, 1, 1, 0,
-            xapp->black(),
-            xapp->black());
+    XSetWindowAttributes attr = {};
+    attr.background_pixel =
+    attr.border_pixel = xapp->black();
+    attr.override_redirect = True;
+    unsigned long mask = CWBackPixel | CWBorderPixel | CWOverrideRedirect;
+    Window xid = XCreateWindow(xapp->display(), xroot, -1, -1, 1, 1, 0, 0,
+                               InputOutput, CopyFromParent, mask, &attr);
 
     XSetSelectionOwner(xapp->display(), wmSx, xid, timestamp);
 
     if (XGetSelectionOwner(xapp->display(), wmSx) != xid)
         die(1, _("Failed to become the owner of the %s selection"), wmSx.str());
 
-    if (current_wm != None) {
+    if (current_wm) {
         XEvent event;
         msg(_("Waiting to replace the old window manager"));
         do {
@@ -561,25 +563,21 @@ void TileMenu::updatePopup() {
         return;
     }
 
-    addItem(_("Left Half"),    -2, actionTileLeft, nullptr, "tileleft");
-    addItem(_("Right Half"),   -2, actionTileRight, nullptr, "tileright");
-    addItem(_("Top Half"),     -2, actionTileTop, nullptr, "tiletop");
-    addItem(_("Bottom Half"),  -2, actionTileBottom, nullptr, "tilebottom");
+    addItem(_("Left Half"),    -2, gKeyWinTileLeft.name, actionTileLeft, "tileleft");
+    addItem(_("Right Half"),   -2, gKeyWinTileRight.name, actionTileRight, "tileright");
+    addItem(_("Top Half"),     -2, gKeyWinTileTop.name, actionTileTop, "tiletop");
+    addItem(_("Bottom Half"),  -2, gKeyWinTileBottom.name, actionTileBottom, "tilebottom");
     addSeparator();
-    addItem(_("Top Left"),     -2, actionTileTopLeft, nullptr,
-            "tiletopleft");
-    addItem(_("Top Right"),    -2, actionTileTopRight, nullptr,
-            "tiletopright");
-    addItem(_("Bottom Left"),  -2, actionTileBottomLeft, nullptr,
-            "tilebottomleft");
-    addItem(_("Bottom Right"), -2, actionTileBottomRight, nullptr,
-            "tilebottomright");
-    addItem(_("Center"),       -2, actionTileCenter, nullptr, "tilecenter");
+    addItem(_("Top Left"),     -2, gKeyWinTileTopLeft.name, actionTileTopLeft, "tiletopleft");
+    addItem(_("Top Right"),    -2, gKeyWinTileTopRight.name, actionTileTopRight, "tiletopright");
+    addItem(_("Bottom Left"),  -2, gKeyWinTileBottomLeft.name, actionTileBottomLeft, "tilebottomleft");
+    addItem(_("Bottom Right"), -2, gKeyWinTileBottomRight.name, actionTileBottomRight, "tilebottomright");
+    addItem(_("Center"),       -2, gKeyWinTileCenter.name, actionTileCenter, "tilecenter");
     addSeparator();
     addItem(_("T_ile Horizontally"), -2,
-            KEY_NAME(gKeySysTileHorizontal), actionTileHorizontal);
+            gKeySysTileHorizontal.name, actionTileHorizontal);
     addItem(_("Tile _Vertically"), -2,
-            KEY_NAME(gKeySysTileVertical), actionTileVertical);
+            gKeySysTileVertical.name, actionTileVertical);
 }
 
 YMenu* YWMApp::getWindowMenu() {
@@ -1183,6 +1181,16 @@ static void showExtensions() {
 
 static int restartWM(const char* displayName, const char* overrideTheme) {
     Display* display = XOpenDisplay(displayName);
+    if (display) {
+        char name[32];
+        snprintf(name, sizeof name, "WM_S%d", DefaultScreen(display));
+        Atom selatom = XInternAtom(display, name, False);
+        Window owner = XGetSelectionOwner(display, selatom);
+        if (owner == None) {
+            XCloseDisplay(display);
+            return -1;
+        }
+    }
     if (display) {
         if (nonempty(overrideTheme)) {
             WMConfig::setDefaultTheme(overrideTheme);
@@ -1783,8 +1791,13 @@ int main(int argc, char **argv) {
         }
     }
 
-    if (restart_wm)
-        return restartWM(displayName, overrideTheme);
+    if (restart_wm) {
+        int r = restartWM(displayName, overrideTheme);
+        if (r >= 0)
+            return r;
+        if (fork())
+            return 0;
+    }
     if (rewrite_prefs)
         return WMConfig::rewritePrefs(wmapp_preferences, configFile);
 
