@@ -15,7 +15,7 @@ public:
     virtual void paintFocus(Graphics &g, const YRect &r);
     virtual void handleButton(const XButtonEvent &button);
     virtual void handleClick(const XButtonEvent &up, int count);
-    virtual void handleBeginDrag(const XButtonEvent &down, const XMotionEvent &motion);
+    virtual bool handleBeginDrag(const XButtonEvent &down, const XMotionEvent &motion);
     virtual void handleVisibility(const XVisibilityEvent& visibility);
     virtual void handleExpose(const XExposeEvent& expose) {}
     virtual void configure(const YRect2 &r);
@@ -28,7 +28,6 @@ public:
     ref<YPixmap> getPixmap(int pn) const;
     YFrameWindow *getFrame() const;
     YFrameTitleBar *titlebar() const;
-    bool onRight() const;
     char getKind() const { return fKind; }
 
 private:
