@@ -31,6 +31,7 @@
 #include "aworkspaces.h"
 #include "yxtray.h"
 #include "prefs.h"
+#include "yprefs.h"
 #include "wpixmaps.h"
 #include "aapm.h"
 
@@ -45,7 +46,7 @@ EdgeTrigger::EdgeTrigger(TaskBar *owner):
     fHideOrShow(Hide)
 {
     setStyle(wsOverrideRedirect | wsInputOnly);
-    setPointer(YXApplication::leftPointer);
+    setPointer(YWMApp::leftPointer);
     setDND(enabled());
     setTitle("IceEdge");
 }
@@ -74,7 +75,7 @@ void EdgeTrigger::show(bool enable) {
         YWindow::show();
     } else {
         YWindow::hide();
-        if (enabled) {
+        if (enabled && !fTaskBar->isCollapsed()) {
             startTimer();
         }
     }
@@ -83,7 +84,8 @@ void EdgeTrigger::show(bool enable) {
 void EdgeTrigger::handleCrossing(const XCrossingEvent &crossing) {
     if (crossing.type == EnterNotify /* && crossing.mode != NotifyNormal */) {
         unsigned long last = YWindow::getLastEnterNotifySerial();
-        if (crossing.serial != last && crossing.serial != last + 1) {
+        if ((crossing.serial != last && crossing.serial != last + 1) ||
+            (crossing.mode == NotifyNormal && crossing.detail == NotifyAncestor)) {
             MSG(("enter notify %d %d", crossing.mode, crossing.detail));
             startTimer(Show);
         }
@@ -133,13 +135,13 @@ TaskBar::TaskBar(IApp *app, YWindow *aParent, YActionListener *wmActionListener,
     fFullscreen(false),
     fIsCollapsed(false),
     fMenuShown(false),
-    fNeedRelayout(false),
+    fNeedRelayout(true),
     fButtonUpdate(false),
     fWorkspacesUpdate(false)
 {
     taskBar = this;
 
-    addStyle(wsNoExpose);
+    addStyle(wsDesktopAware | wsTakeFocus | wsNoExpose);
     setWinHintsHint(WinHintsSkipFocus |
                     WinHintsSkipWindowMenu |
                     WinHintsSkipTaskBar);
@@ -162,7 +164,7 @@ TaskBar::TaskBar(IApp *app, YWindow *aParent, YActionListener *wmActionListener,
        MWM_HINTS_FUNCTIONS | MWM_HINTS_DECORATIONS,
        MWM_FUNC_MOVE));
     setFrameState(NormalState);
-    setPointer(YXApplication::leftPointer);
+    setPointer(YWMApp::leftPointer);
     setDND(true);
 
     fEdgeTrigger = new EdgeTrigger(this);
@@ -272,7 +274,7 @@ void TaskBar::initApplets() {
         fNetStatus = nullptr;
 
     if (taskBarShowClock)
-        fClock = new YClock(smActionListener, this, this);
+        fClock = new ClockSet(smActionListener, this, this);
     else
         fClock = nullptr;
 
@@ -303,8 +305,13 @@ void TaskBar::initApplets() {
         fCollapseButton = new ObjectButton(this, actionCollapseTaskbar);
         if (fCollapseButton) {
             fCollapseButton->setWinGravity(StaticGravity);
-            fCollapseButton->setText(">");
-            fCollapseButton->setImage(taskbarCollapseImage);
+            ref<YImage> image = leftToRight
+                              ? taskbarCollapseImage : taskbarExpandImage;
+            if (image != null) {
+                fCollapseButton->setImage(image);
+            } else {
+                fCollapseButton->setText(leftToRight ? ">" : "<");
+            }
             fCollapseButton->setActionListener(this);
             fCollapseButton->setToolTip(_("Hide Taskbar"));
             fCollapseButton->setTitle("Collapse");
@@ -318,7 +325,7 @@ void TaskBar::initApplets() {
         fMailBoxControl = nullptr;
 
     if (configKeyboards.nonempty()) {
-        fKeyboardStatus = new KeyboardStatus(this, this);
+        fKeyboardStatus = new KeyboardStatus(wmapp, this, this);
     } else
         fKeyboardStatus = nullptr;
 
@@ -402,8 +409,10 @@ void TaskBar::initApplets() {
                                   this, trayDrawBevel);
         fDesktopTray->setTitle("SystemTray");
         fDesktopTray->relayout();
-    } else
+    } else {
         fDesktopTray = nullptr;
+        updateLocation();
+    }
 
     if (fCollapseButton) {
         fCollapseButton->raise();
@@ -412,7 +421,6 @@ void TaskBar::initApplets() {
 
 void TaskBar::trayChanged() {
     relayout();
-    //    updateLayout();
 }
 
 struct LayoutInfo {
@@ -462,8 +470,11 @@ void TaskBar::updateLayout(unsigned &size_w, unsigned &size_h) {
         wlist.append(nw);
     }
 
-    nw = LayoutInfo( fClock, Over, Top, Same, Keep, 2, 2 );
-    wlist.append(nw);
+    if (fClock) {
+        for (YClock* clock : *fClock) {
+            wlist += LayoutInfo( clock, Over, Top, Same, Keep, 2, 2 );
+        }
+    }
 
     if (taskBarShowMailboxStatus) {
         for (auto m = fMailBoxControl->iterator(); ++m; ) {
@@ -577,7 +588,11 @@ void TaskBar::updateLayout(unsigned &size_w, unsigned &size_h) {
 
             right[wlist[i].row] -= ww + wlist[i].pre + wlist[i].post;
         }
-        wlist[i].w->setGeometry(YRect(xx, yy, ww, hh));
+        YRect r(xx, yy, ww, hh);
+        if (rightToLeft) {
+            r.xx = w - r.xx - r.ww;
+        }
+        wlist[i].w->setGeometry(r);
         if (wlist[i].show)
             wlist[i].w->show();
     }
@@ -585,28 +600,29 @@ void TaskBar::updateLayout(unsigned &size_w, unsigned &size_h) {
     wlist.clear();
     /* ----------------------------------------------------------------- */
 
-    if (taskBarShowWindows) {
-        if (fTasks) {
-            fTasks->hide();
-            fTasks->setGeometry(YRect(left[0],
-                                      y[0],
-                                      max(0, right[0] - left[0]),
-                                      h[0]));
-            fTasks->show();
-            fTasks->relayout();
+    if (taskBarShowWindows && fTasks) {
+        fTasks->hide();
+        YRect r(left[0], y[0], unsigned(max(1, right[0] - left[0])), h[0]);
+        if (rightToLeft) {
+            r.xx = w - r.xx - r.ww;
         }
+        fTasks->setGeometry(r);
+        fTasks->show();
+        fTasks->relayout();
     }
     if (fAddressBar) {
         int row = taskBarDoubleHeight;
-
-        fAddressBar->setGeometry(YRect(left[row],
-                                       y[row] + 2,
-                                       max(0, right[row] - left[row]),
-                                       h[row] - 4));
-        fAddressBar->raise();
+        YRect r(left[row], y[row] + 2,
+                max(1U, unsigned(right[row] - left[row])), h[row] - 4);
+        if (rightToLeft) {
+            r.xx = w - r.xx - r.ww;
+        }
+        fAddressBar->setGeometry(r);
         if (::showAddressBar) {
-            if (taskBarDoubleHeight || !taskBarShowWindows)
+            if (taskBarDoubleHeight || !taskBarShowWindows) {
+                fAddressBar->raise();
                 fAddressBar->show();
+            }
         }
     }
 
@@ -616,8 +632,10 @@ void TaskBar::updateLayout(unsigned &size_w, unsigned &size_h) {
 
 void TaskBar::relayoutNow() {
     if (fUpdates.nonempty()) {
-        for (YFrameWindow* frame : fUpdates) {
-            frame->updateAppStatus();
+        for (int i = fUpdates.getCount(); --i >= 0; ) {
+            if (i < fUpdates.getCount() && fUpdates[i]) {
+                fUpdates[i]->updateAppStatus();
+            }
         }
         fUpdates.clear();
     }
@@ -638,9 +656,14 @@ void TaskBar::relayoutNow() {
     }
 }
 
-void TaskBar::updateFullscreen(bool fullscreen) {
-    if (fFullscreen != fullscreen && getFrame()) {
+void TaskBar::updateFullscreen() {
+    bool fullscreen = manager->getFocus() &&
+                      manager->getFocus()->isFullscreen() &&
+                      manager->getFocus()->visible();
+    if (fFullscreen != fullscreen) {
         fFullscreen = fullscreen;
+        if (fFullscreen == false)
+            updateWinLayer();
         fEdgeTrigger->show((fFullscreen | fIsHidden) && !fIsCollapsed);
     }
 }
@@ -648,6 +671,9 @@ void TaskBar::updateFullscreen(bool fullscreen) {
 void TaskBar::updateLocation() {
     fNeedRelayout = false;
 
+    if (getFrame() == nullptr) {
+        showBar();
+    }
     if (fIsHidden && !fIsCollapsed) {
         if (getFrame() && visible())
             getFrame()->wmHide();
@@ -664,10 +690,14 @@ void TaskBar::updateLocation() {
 
     if (taskBarWidthPercentage < 100) {
         w = (dw * taskBarWidthPercentage + 50) / 100;
-        if (strcmp(taskBarJustify, "right") == 0)
-            x = dx + (dw - w);
-        if (strcmp(taskBarJustify, "center") == 0)
-            x = dx + (dw - w)/2;
+        if (nonempty(taskBarJustify)) {
+            if (strcmp(taskBarJustify, "left") == 0)
+                x = dx;
+            else if (strcmp(taskBarJustify, "right") == 0)
+                x = dx + (dw - w);
+            else if (strcmp(taskBarJustify, "center") == 0)
+                x = dx + (dw - w)/2;
+        }
     }
 
     updateLayout(w, h);
@@ -684,19 +714,19 @@ void TaskBar::updateLocation() {
             w = h = 0;
         }
 
-        x = dx + (dw - w);
+        x = rightToLeft ? dx : dx + (dw - w);
     }
 
     int by = taskBarAtTop ? dy : dy + dh - 1;
 
-    fEdgeTrigger->setGeometry(YRect(x, by, w, 1));
+    fEdgeTrigger->setGeometry(YRect(x, by, w, 1U));
 
     int y = taskBarAtTop ? dy : dy + dh - h;
 
     if ( !fIsHidden || fIsCollapsed) {
         if (getFrame()) {
             if (geometry() != YRect(x, y, w, h)) {
-                XConfigureRequestEvent conf;
+                XConfigureRequestEvent conf = {};
                 conf.type = ConfigureRequest;
                 conf.window = handle();
                 conf.x = x;
@@ -774,6 +804,9 @@ void TaskBar::handleCrossing(const XCrossingEvent &crossing) {
         } else {
             if (ahwm_hack) {
                 fEdgeTrigger->startTimer();
+                if (manager->getFocus() == getFrame()) {
+                    manager->focusLastWindow();
+                }
             }
         }
     }
@@ -785,8 +818,17 @@ void TaskBar::paint(Graphics &g, const YRect& r) {
          fGradient->width() != width() ||
          fGradient->height() != height()))
     {
-        int gradientHeight = height() / (1 + taskBarDoubleHeight);
-        fGradient = taskbackPixbuf->scale(width(), gradientHeight);
+        if (taskBarDoubleHeight == false) {
+            fGradient = taskbackPixbuf->scale(width(), height());
+        } else {
+            ref<YImage> s = taskbackPixbuf->scale(width(), (height() + 1) / 2);
+            ref<YPixmap> p = YPixmap::create(width(), height(), depth());
+            Graphics g(p);
+            g.clear();
+            g.copyImage(s, 0, 0);
+            g.copyImage(s, 0, height() - s->height());
+            fGradient = YImage::createFromPixmap(p);
+        }
     }
 
     g.setColor(taskBarBg);
@@ -826,7 +868,7 @@ bool TaskBar::handleKey(const XKeyEvent &key) {
 void TaskBar::handleButton(const XButtonEvent &button) {
     if ((button.type == ButtonRelease) &&
         (button.button == 1 || button.button == 3) &&
-        (BUTTON_MODMASK(button.state) == Button1Mask + Button3Mask))
+        xapp->isButton(button.state, Button1Mask + Button3Mask))
     {
         windowList->showFocused(button.x_root, button.y_root);
     }
@@ -855,7 +897,7 @@ void TaskBar::handleClick(const XButtonEvent &up, int count) {
     } else if (up.button == 2) {
         windowList->showFocused(up.x_root, up.y_root);
     } else {
-        if (up.button == 3 && count == 1 && IS_BUTTON(up.state, Button3Mask)) {
+        if (up.button == 3 && count == 1 && xapp->isButton(up.state, Button3Mask)) {
             contextMenu(up.x_root, up.y_root);
         }
     }
@@ -865,31 +907,25 @@ void TaskBar::handleEndDrag(const XButtonEvent &/*down*/, const XButtonEvent &/*
     xapp->releaseEvents();
 }
 void TaskBar::handleDrag(const XButtonEvent &/*down*/, const XMotionEvent &motion) {
-    int newPosition = 0;
+    bool newPosition = false;
 
-    xapp->grabEvents(this, YXApplication::movePointer.handle(),
+    xapp->grabEvents(this, YWMApp::movePointer,
                          ButtonPressMask |
                          ButtonReleaseMask |
                          PointerMotionMask);
 
 
     if (motion.y_root < int(desktop->height() / 2))
-        newPosition = 1;
+        newPosition = true;
 
     if (taskBarAtTop != newPosition) {
         taskBarAtTop = newPosition;
-        //setPosition(x(), taskBarAtTop ? -1 : int(desktop->height() - height() + 1));
         updateLocation();
-        //repaint();
-        //manager->updateWorkArea();
     }
 }
 
 void TaskBar::popupStartMenu() {
     if (fApplications) {
-        /*requestFocus();
-         fApplications->requestFocus();
-         fApplications->setFocus();*/
         popOut();
         fApplications->popupMenu();
     }
@@ -903,21 +939,21 @@ void TaskBar::popupWindowListMenu() {
 }
 
 bool TaskBar::autoTimer(bool doShow) {
-    MSG(("hide taskbar"));
+    MSG(("taskbar timer show=%s full=%s hidden=%s",
+        boolstr(doShow), boolstr(fFullscreen), boolstr(fIsHidden)));
+    updateFullscreen();
     if (fFullscreen && doShow && taskBarFullscreenAutoShow) {
         fIsHidden = false;
-        getFrame()->focus();
-        manager->switchFocusTo(getFrame(), true);
-        manager->updateFullscreenLayer();
+        obtainFocus();
     }
     if (taskBarAutoHide) {
-        fIsHidden = !doShow && !hasPopup();
+        fIsHidden = !doShow && !hasPopup() && manager->getFocus() != getFrame();
         if (taskBarDoubleHeight == false && taskBarShowWindows) {
             fIsHidden &= !(addressBar() && addressBar()->visible());
         }
         updateLocation();
     }
-    return fIsHidden == doShow;
+    return fIsHidden == doShow && taskBarAutoHide && !fIsCollapsed;
 }
 
 void TaskBar::popOut() {
@@ -925,9 +961,10 @@ void TaskBar::popOut() {
         handleCollapseButton();
     }
     if (taskBarAutoHide) {
+        bool hidden = fIsHidden;
         fIsHidden = false;
         updateLocation();
-        fIsHidden = taskBarAutoHide;
+        fIsHidden = hidden;
         if (fEdgeTrigger) {
             MSG(("start hide 4"));
             fEdgeTrigger->startTimer();
@@ -938,13 +975,11 @@ void TaskBar::popOut() {
 
 void TaskBar::showBar() {
     if (getFrame() == nullptr) {
-        manager->manageClient(handle());
-        updateWinLayer();
+        manager->manageClient(this);
         if (getFrame()) {
             getFrame()->setAllWorkspaces();
             if (enableAddressBar && ::showAddressBar && taskBarDoubleHeight)
-                getFrame()->activate(true);
-            updateLocation();
+                obtainFocus();
             parent()->setTitle("TaskBarFrame");
             getFrame()->updateLayer();
         }
@@ -958,9 +993,16 @@ void TaskBar::actionPerformed(YAction action, unsigned int modifiers) {
 void TaskBar::handleCollapseButton() {
     fIsCollapsed = !fIsCollapsed;
     if (fCollapseButton) {
-        fCollapseButton->setText(fIsCollapsed ? "<": ">");
-        fCollapseButton->setImage(fIsCollapsed ? taskbarExpandImage : taskbarCollapseImage);
-        fCollapseButton->setToolTip(fIsCollapsed ? _("Show Taskbar") : _("Hide Taskbar"));
+        ref<YImage> image = (leftToRight == fIsCollapsed)
+                          ? taskbarExpandImage : taskbarCollapseImage;
+        const char* text = (leftToRight == fIsCollapsed) ? "<" : ">";
+        const char* ttip = fIsCollapsed ? _("Show Taskbar") : _("Hide Taskbar");
+        if (image != null) {
+            fCollapseButton->setImage(image);
+        } else {
+            fCollapseButton->setText(text);
+        }
+        fCollapseButton->setToolTip(ttip);
         fCollapseButton->repaint();
     }
 
@@ -977,8 +1019,9 @@ void TaskBar::handlePopDown(YPopupWindow * /*popup*/) {
 }
 
 void TaskBar::configure(const YRect2& r) {
-    if (r.resized() && 1 < r.width()) {
+    if (r.resized() && 1 < r.width() && !fIsCollapsed) {
         repaint();
+        clearWindow();
     }
     updateWMHints();
 }
@@ -1028,11 +1071,18 @@ void TaskBar::relayoutTray() {
 
 void TaskBar::showAddressBar() {
     popOut();
-    if (fAddressBar != nullptr)
+    if (fAddressBar) {
+        obtainFocus();
         fAddressBar->showNow();
+    }
 }
 
-void TaskBar::setWorkspaceActive(long workspace, bool active) {
+void TaskBar::obtainFocus() {
+    getFrame()->focus();
+    manager->switchFocusTo(getFrame(), true);
+}
+
+void TaskBar::setWorkspaceActive(int workspace, bool active) {
     if (taskBarShowWorkspaces && fWorkspaces) {
         YDimension dim(fWorkspaces->dimension());
         fWorkspaces->setPressed(workspace, active);
@@ -1042,9 +1092,10 @@ void TaskBar::setWorkspaceActive(long workspace, bool active) {
     }
 }
 
-void TaskBar::workspacesRepaint() {
+void TaskBar::workspacesRepaint(int workspace) {
     if (taskBarShowWorkspaces && fWorkspaces) {
         fWorkspacesUpdate = true;
+        fWorkspaces->repaintWorkspace(workspace);
     }
 }
 
