@@ -7,6 +7,7 @@
 #include "ypaint.h"
 #include "yxapp.h"
 #include "yprefs.h"
+#include "yfontbase.h"
 #include "ascii.h"
 #include "intl.h"
 #include <stdlib.h>
@@ -19,9 +20,6 @@
 #include <wctype.h>
 #endif
 
-#define utf8ellipsis "\xe2\x80\xa6"
-const unsigned utf32ellipsis = 0x2026;
-
 static inline Display* display()  { return xapp->display(); }
 
 /******************************************************************************/
@@ -29,7 +27,7 @@ static inline Display* display()  { return xapp->display(); }
 Graphics::Graphics(YWindow & window,
                    unsigned long vmask, XGCValues * gcv):
     fDrawable(window.handle()),
-    fColor(), fFont(null),
+    fColor(), fFont(),
     fPicture(None),
     xOrigin(0), yOrigin(0)
 {
@@ -44,10 +42,10 @@ Graphics::Graphics(YWindow & window,
 
 Graphics::Graphics(YWindow & window):
     fDrawable(window.handle()),
-    fColor(), fFont(null),
+    fColor(), fFont(),
     fPicture(None),
     xOrigin(0), yOrigin(0)
- {
+{
     rWidth = window.width();
     rHeight = window.height();
     rDepth = (window.depth() ? window.depth() : xapp->depth());
@@ -58,12 +56,12 @@ Graphics::Graphics(YWindow & window):
 #endif
 }
 
-Graphics::Graphics(ref<YPixmap> pixmap, int x_org, int y_org):
+Graphics::Graphics(ref<YPixmap> pixmap):
     fDrawable(pixmap->pixmap()),
-    fColor(), fFont(null),
+    fColor(), fFont(),
     fPicture(None),
-    xOrigin(x_org), yOrigin(y_org)
- {
+    xOrigin(0), yOrigin(0)
+{
     rWidth = pixmap->width();
     rHeight = pixmap->height();
     rDepth = pixmap->depth();
@@ -77,7 +75,7 @@ Graphics::Graphics(ref<YPixmap> pixmap, int x_org, int y_org):
 Graphics::Graphics(Drawable drawable, unsigned w, unsigned h, unsigned depth,
                    unsigned long vmask, XGCValues * gcv):
     fDrawable(drawable),
-    fColor(), fFont(null),
+    fColor(), fFont(),
     fPicture(None),
     xOrigin(0), yOrigin(0),
     rWidth(w), rHeight(h), rDepth(depth)
@@ -90,7 +88,7 @@ Graphics::Graphics(Drawable drawable, unsigned w, unsigned h, unsigned depth,
 
 Graphics::Graphics(Drawable drawable, unsigned w, unsigned h, unsigned depth):
     fDrawable(drawable),
-    fColor(), fFont(null),
+    fColor(), fFont(),
     fPicture(None),
     xOrigin(0), yOrigin(0),
     rWidth(w), rHeight(h), rDepth(depth)
@@ -226,18 +224,22 @@ void Graphics::drawLine(int x1, int y1, int x2, int y2) {
 }
 
 void Graphics::drawSegments(XSegment *segments, int n) {
-    for (int i = 0; i < n; i++) {
-        segments[i].x1 -= xOrigin;
-        segments[i].y1 -= yOrigin;
-        segments[i].x2 -= xOrigin;
-        segments[i].y2 -= yOrigin;
+    if (xOrigin | yOrigin) {
+        for (int i = 0; i < n; i++) {
+            segments[i].x1 -= xOrigin;
+            segments[i].y1 -= yOrigin;
+            segments[i].x2 -= xOrigin;
+            segments[i].y2 -= yOrigin;
+        }
     }
     XDrawSegments(display(), drawable(), gc, segments, n);
-    for (int i = 0; i < n; i++) {
-        segments[i].x1 += xOrigin;
-        segments[i].y1 += yOrigin;
-        segments[i].x2 += xOrigin;
-        segments[i].y2 += yOrigin;
+    if (xOrigin | yOrigin) {
+        for (int i = 0; i < n; i++) {
+            segments[i].x1 += xOrigin;
+            segments[i].y1 += yOrigin;
+            segments[i].x2 += xOrigin;
+            segments[i].y2 += yOrigin;
+        }
     }
 }
 
@@ -247,14 +249,18 @@ void Graphics::drawRect(int x, int y, unsigned width, unsigned height) {
 }
 
 void Graphics::drawRects(XRectangle *rects, unsigned n) {
-    for (unsigned i = 0; i < n; i++) {
-        rects[i].x -= xOrigin;
-        rects[i].y -= yOrigin;
+    if (xOrigin | yOrigin) {
+        for (unsigned i = 0; i < n; i++) {
+            rects[i].x -= xOrigin;
+            rects[i].y -= yOrigin;
+        }
     }
     XDrawRectangles(display(), drawable(), gc, rects, int(n));
-    for (unsigned i = 0; i < n; i++) {
-        rects[i].x += xOrigin;
-        rects[i].y += yOrigin;
+    if (xOrigin | yOrigin) {
+        for (unsigned i = 0; i < n; i++) {
+            rects[i].x += xOrigin;
+            rects[i].y += yOrigin;
+        }
     }
 }
 
@@ -277,7 +283,14 @@ void Graphics::drawChars(const char *data, int offset, int len, int x, int y) {
 }
 
 void Graphics::drawString(int x, int y, char const * str) {
-    drawChars(str, 0, int(strlen(str)), x, y);
+    if (fFont != null)
+        fFont->drawGlyphs(*this, x, y, str, int(strlen(str)));
+}
+
+void Graphics::drawChars(wchar_t* data, int offset, int len, int x, int y)
+{
+    if (fFont != null)
+        fFont->drawGlyphs(*this, x, y, data + offset, len);
 }
 
 /**
@@ -285,74 +298,10 @@ void Graphics::drawString(int x, int y, char const * str) {
  * ... character.
  */
 void Graphics::drawStringEllipsis(int x, int y, const char *str, int maxWidth) {
-    const int len(strlen(str));
-
-    if (fFont == null || fFont->textWidth(str, len) <= maxWidth) {
-        drawChars(str, 0, len, x, y);
-        return;
-    }
-
-    auto ellipsis = showEllipsis && fFont->supports(utf32ellipsis) ?
-            utf8ellipsis : "...";
-
-    if (showEllipsis)
-        maxWidth -= fFont->textWidth(ellipsis, 3);
-
-    int rawPos(0), drawPos(0), trimLen(0), trimWid(0);
-
-    if (maxWidth > 0) {
-        while (rawPos < len) {
-            int glyphLen, glyphWidth;
-#ifdef CONFIG_I18N
-            wchar_t wc;
-            if (multiByte) {
-                glyphLen = mbtowc(&wc, str + rawPos, size_t(len - rawPos));
-                if (glyphLen < 1) { // bad things
-                    rawPos++;
-                    continue;
-                }
-                glyphWidth = fFont->textWidth(str + rawPos, glyphLen);
-            } else
-#endif
-            {
-                glyphLen = 1;
-                glyphWidth = fFont->textWidth(str + rawPos, 1);
-            }
-
-            if (drawPos + glyphWidth >= maxWidth)
-                break;
-            bool isSpace =
-#ifdef CONFIG_I18N
-                multiByte ?
-                    !iswprint(wc) :
-#endif
-                    (1 == glyphLen && ASCII::isWhiteSpace(str[rawPos]));
-
-            if (isSpace)
-            {
-                trimLen += glyphLen;
-                trimWid += glyphWidth;
-            } else {
-                trimLen = trimWid = 0;
-            }
-
-            rawPos += glyphLen;
-            drawPos += glyphWidth;
-        }
-        // chop off trailing whitespace
-        rawPos -= trimLen;
-        drawPos -= trimWid;
-    }
-
-    if (rawPos > 0)
-        drawChars(str, 0, rawPos, x, y);
-
-    if (showEllipsis && rawPos < len) {
-        drawChars(ellipsis, 0, 3, x + drawPos, y);
-    }
+    if (fFont != null)
+        fFont->drawGlyphs(*this, x, y, str, int(strlen(str)), maxWidth);
 }
 
-
 void Graphics::drawCharUnderline(int x, int y, const char *str, int charPos) {
 /// TODO #warning "FIXME: don't mess with multibyte here, use a wide char"
     int left = 0; //fFont ? fFont->textWidth(str, charPos) : 0;
@@ -397,32 +346,40 @@ void Graphics::drawCharUnderline(int x, int y, const char *str, int charPos) {
         drawLine(x + left, y + 2, x + right, y + 2);
 }
 
-void Graphics::drawStringMultiline(int x, int y, const char *str) {
-    unsigned const tx(x + fFont->multilineTabPos(str));
+void Graphics::drawStringMultiline(const char* str, int x, int y, unsigned w) {
+    const int tabpos = fFont->multilineTabPos(str);
+    const int tabx = tabpos ? x + tabpos : x;
 
-    for (const char * end(strchr(str, '\n')); end;
-         str = end + 1, end = strchr(str, '\n')) {
-        int const len(int(end - str));
-        const char* tab(static_cast<const char *>(memchr(str, '\t', len)));
+    while (*str) {
+        const char* nln = strchr(str, '\n');
+        const char* end = nln ? nln : str + strlen(str);
+        const int len = int(end - str);
+        const char* tab = static_cast<const char *>(memchr(str, '\t', len));
 
-        if (tab) {
-            drawChars(str, 0, int(tab - str), x, y);
-            drawChars(tab + 1, 0, int(end - tab) - 1, int(tx), y);
+        if (leftToRight) {
+            if (tab) {
+                const char* ing = tab + 1;
+                drawChars(str, 0, int(tab - str), x, y);
+                drawChars(ing, 0, int(end - ing), tabx, y);
+            } else {
+                drawChars(str, 0, int(end - str), x, y);
+            }
+        } else if (rightToLeft) {
+            if (tab) {
+                const char* ing = tab + 1;
+                int sw = fFont->textWidth(str, int(tab - str));
+                int iw = fFont->textWidth(ing, int(end - ing));
+                drawChars(str, 0, int(tab - str), x + w - sw, y);
+                drawChars(ing, 0, int(end - ing), x + w - tabpos - iw, y);
+            } else {
+                int sw = fFont->textWidth(str, int(end - str));
+                drawChars(str, 0, int(end - str), x + w - sw, y);
+            }
         }
-        else
-            drawChars(str, 0, int(end - str), x, y);
 
-        y+= fFont->height();
+        y += fFont->height();
+        str = *end ? end + 1 : end;
     }
-
-    const char * tab(strchr(str, '\t'));
-
-    if (tab) {
-        drawChars(str, 0, int(tab - str), x, y);
-        drawChars(tab + 1, 0, int(strlen(tab + 1)), int(tx), y);
-    }
-    else
-        drawChars(str, 0, int(strlen(str)), x, y);
 }
 
 /******************************************************************************/
@@ -432,15 +389,79 @@ void Graphics::fillRect(int x, int y, unsigned width, unsigned height) {
                    x - xOrigin, y - yOrigin, width, height);
 }
 
+void Graphics::fillRect(int x, int y, unsigned width, unsigned height,
+                        unsigned roundedCornerRadius)
+{
+    unsigned rounding = min(roundedCornerRadius, min(width, height) / 2);
+    if (rounding == 0) {
+        fillRect(x, y, width, height);
+    } else {
+        XArc arc[4];
+
+        x -= xOrigin;
+        y -= yOrigin;
+
+        arc[0].x = x;
+        arc[0].y = y;
+        arc[0].width = 2 * rounding;
+        arc[0].height = 2 * rounding;
+        arc[0].angle1 = 90 * 64;
+        arc[0].angle2 = 90 * 64;
+
+        arc[1] = arc[0];
+        arc[1].y = y + height - 2 * rounding - 1;
+        arc[1].angle1 = 180 * 64;
+
+        arc[2] = arc[1];
+        arc[2].x = x + width - 2 * rounding - 1;
+        arc[2].angle1 = 270 * 64;
+
+        arc[3] = arc[2];
+        arc[3].y = y;
+        arc[3].angle1 = 0 * 64;
+
+        XFillArcs(display(), drawable(), gc, arc, 4);
+
+        XRectangle rec[3];
+        int n = 0;
+        int t1 = arc[0].x + rounding;
+        int t2 = arc[2].x + rounding;
+        if (t1 <= t2) {
+            rec[n].x = t1;
+            rec[n].width = t2 - t1 + 1;
+            rec[n].y = arc[0].y;
+            rec[n].height = height;
+            ++n;
+        }
+        int u1 = arc[0].y + rounding;
+        int u2 = arc[1].y + rounding;
+        if (u1 <= u2) {
+            rec[n].x = x;
+            rec[n].width = rounding;
+            rec[n].y = u1;
+            rec[n].height = u2 - u1 + 1;
+            ++n;
+            rec[n] = rec[n - 1];
+            rec[n].x += width - rounding;
+            ++n;
+        }
+        XFillRectangles(display(), drawable(), gc, rec, n);
+    }
+}
+
 void Graphics::fillRects(XRectangle *rects, int n) {
-    for (int i = 0; i < n; i++) {
-        rects[i].x -= xOrigin;
-        rects[i].y -= yOrigin;
+    if (xOrigin | yOrigin) {
+        for (int i = 0; i < n; i++) {
+            rects[i].x -= xOrigin;
+            rects[i].y -= yOrigin;
+        }
     }
     XFillRectangles(display(), drawable(), gc, rects, n);
-    for (int i = 0; i < n; i++) {
-        rects[i].x += xOrigin;
-        rects[i].y += yOrigin;
+    if (xOrigin | yOrigin) {
+        for (int i = 0; i < n; i++) {
+            rects[i].x += xOrigin;
+            rects[i].y += yOrigin;
+        }
     }
 }
 
@@ -449,14 +470,18 @@ void Graphics::fillPolygon(XPoint *points, int n, int shape,
 {
     int n1 = (mode == CoordModeOrigin) ? n : 1;
 
-    for (int i = 0; i < n1; i++) {
-        points[i].x -= xOrigin;
-        points[i].y -= yOrigin;
+    if (xOrigin | yOrigin) {
+        for (int i = 0; i < n1; i++) {
+            points[i].x -= xOrigin;
+            points[i].y -= yOrigin;
+        }
     }
     XFillPolygon(display(), drawable(), gc, points, n, shape, mode);
-    for (int i = 0; i < n1; i++) {
-        points[i].x += xOrigin;
-        points[i].y += yOrigin;
+    if (xOrigin | yOrigin) {
+        for (int i = 0; i < n1; i++) {
+            points[i].x += xOrigin;
+            points[i].y += yOrigin;
+        }
     }
 }
 
@@ -467,10 +492,12 @@ void Graphics::fillArc(int x, int y, unsigned width, unsigned height, int a1, in
 
 /******************************************************************************/
 
-void Graphics::setColor(YColor aColor) {
-    fColor = aColor;
-    unsigned long pixel = fColor.pixel();
-    setColorPixel(pixel);
+void Graphics::setColor(const YColor& aColor) {
+    if (aColor) {
+        fColor = aColor;
+        unsigned long pixel = fColor.pixel();
+        setColorPixel(pixel);
+    }
 }
 
 void Graphics::setColorPixel(unsigned long pixel) {
@@ -480,20 +507,18 @@ void Graphics::setColorPixel(unsigned long pixel) {
     XSetForeground(display(), gc, pixel);
 }
 
-void Graphics::setFont(ref<YFont> aFont) {
-    fFont = aFont;
-}
-
 void Graphics::setLineWidth(unsigned width) {
     XGCValues gcv;
     gcv.line_width = int(width);
     XChangeGC(display(), gc, GCLineWidth, &gcv);
 }
 
-void Graphics::setPenStyle(bool dotLine) {
+void Graphics::setPenStyle(bool dotLine, int cap, int join) {
     XGCValues gcv;
-    unsigned long mask = GCLineStyle;
+    unsigned long mask = GCLineStyle | GCCapStyle | GCJoinStyle;
     gcv.line_style = dotLine ? LineOnOffDash : LineSolid;
+    gcv.cap_style = cap;
+    gcv.join_style = join;
 
     if (dotLine) {
         char dashes[] = { 1 };
@@ -502,8 +527,6 @@ void Graphics::setPenStyle(bool dotLine) {
         XSetDashes(display(), gc, dash_offset, dashes, num_dashes);
 
         gcv.line_width = 1;
-        gcv.cap_style = CapButt;
-        gcv.join_style = JoinMiter;
         mask |= GCLineWidth | GCCapStyle | GCJoinStyle;
     }
 
@@ -550,6 +573,13 @@ void Graphics::drawPixmap(ref<YPixmap> pix, int x, int y) {
 
 void Graphics::drawPixmap(ref<YPixmap> pix, int sx, int sy,
                           unsigned w, unsigned h, int dx, int dy) {
+    Picture source = pix->picture(), destin = picture();
+    if (source && destin) {
+        XRenderComposite(display(), PictOpOver, source, None, destin,
+                         0, 0, 0, 0, dx - xOrigin, dy - yOrigin, w, h);
+        return;
+    }
+
     Pixmap pixmap(pix->pixmap(rdepth()));
     if (pixmap == None) {
         tlog("Graphics::%s: attempt to draw pixmap 0x%lx of depth %d with gc of depth %d\n",
@@ -574,9 +604,10 @@ void Graphics::drawMask(ref<YPixmap> pix, int x, int y) {
 void Graphics::drawClippedPixmap(Pixmap pix, Pixmap clip,
                                  int x, int y, unsigned w, unsigned h, int toX, int toY)
 {
-    unsigned long mask =
+    unsigned long mask = GCFunction |
         GCGraphicsExposures | GCClipMask | GCClipXOrigin | GCClipYOrigin;
     XGCValues gcv;
+    gcv.function = GXcopy;
     gcv.graphics_exposures = False;
     gcv.clip_mask = clip;
     gcv.clip_x_origin = toX - xOrigin;
@@ -774,14 +805,18 @@ void Graphics::drawBorderG(int x, int y, unsigned wid, unsigned hei, bool raised
 
 void Graphics::drawLines(XPoint *points, int n, int mode) {
     int n1 = (mode == CoordModeOrigin) ? n : 1;
-    for (int i = 0; i < n1; i++) {
-        points[i].x -= xOrigin;
-        points[i].y -= yOrigin;
+    if (xOrigin | yOrigin) {
+        for (int i = 0; i < n1; i++) {
+            points[i].x -= xOrigin;
+            points[i].y -= yOrigin;
+        }
     }
     XDrawLines(display(), drawable(), gc, points, n, mode);
-    for (int i = 0; i < n1; i++) {
-        points[i].x += xOrigin;
-        points[i].y += yOrigin;
+    if (xOrigin | yOrigin) {
+        for (int i = 0; i < n1; i++) {
+            points[i].x += xOrigin;
+            points[i].y += yOrigin;
+        }
     }
 }
 
@@ -877,11 +912,17 @@ void Graphics::fillPixmap(ref<YPixmap> pixmap, int x, int y,
     if (xpixmap == None)
         return;
 
-    px%= pw; const int pww(px ? pw - px : 0);
-    py%= ph; const int phh(py ? ph - py : 0);
+    px %= pw;
+    py %= ph;
+    if (px < 0)
+        px += pw;
+    if (py < 0)
+        py += ph;
+    const int pww(px ? pw - px : 0);
+    const int phh(py ? ph - py : 0);
 
-    if (px) {
-        if (py)
+    if (0 < px) {
+        if (0 < py)
             XCopyArea(display(), xpixmap, drawable(), gc,
                       px, py, pww, phh, x - xOrigin, y - yOrigin);
 
@@ -890,10 +931,10 @@ void Graphics::fillPixmap(ref<YPixmap> pixmap, int x, int y,
                       px, 0, pww, min(hh, ph), x - xOrigin, yy - yOrigin);
     }
 
-    for (int xx(x + pww), ww(w - pww); ww > 0; xx+= pw, ww-= pw) {
+    for (int xx(x + pww), ww(w - pww); ww > 0; xx += pw, ww -= pw) {
         int const www(min(ww, pw));
 
-        if (py)
+        if (0 < py)
             XCopyArea(display(), xpixmap, drawable(), gc,
                       0, py, www, phh, xx - xOrigin, y - yOrigin);
 
@@ -1161,7 +1202,56 @@ Pixmap GraphicsBuffer::pixmap() {
     return fPixmap;
 }
 
-/******************************************************************************/
-/******************************************************************************/
+void GraphicsBuffer::scroll(int dx, int dy) {
+    if (fPixmap == None || fDim != window()->dimension()) {
+        paint();
+    }
+    else if (dx == 0 && dy == 0) {
+    }
+    else if (abs(dx) >= int(window()->width())
+          || abs(dy) >= int(window()->height()))
+    {
+        paint();
+    }
+    else {
+        XGCValues gcv = {};
+        unsigned long gcvflags = GCGraphicsExposures;
+        gcv.graphics_exposures = False;
+        GC scrollGC = XCreateGC(xapp->display(), fPixmap, gcvflags, &gcv);
+        int ww = int(window()->width());
+        int hh = int(window()->height());
+        int sx, sy, px, py, pw, ph;
+        if (dy > 0) {
+            sy = dy; py = 0; ph = hh - dy;
+        } else {
+            sy = 0; py = -dy; ph = hh + dy;
+        }
+        if (dx > 0) {
+            sx = dx; px = 0; pw = ww - dx;
+        } else {
+            sx = 0; px = -dx; pw = ww + dx;
+        }
+        XCopyArea(xapp->display(), fPixmap, fPixmap, scrollGC,
+                  sx, sy, pw, ph, px, py);
+        XFreeGC(xapp->display(), scrollGC);
+
+        Graphics gfx(fPixmap, ww, hh, window()->depth());
+        YRect rect[2];
+        int count = 0;
+        if (dy > 0) rect[count++] = { 0, ph, ww, dy };
+        if (dy < 0) rect[count++] = { 0, 0, ww, -dy };
+        if (dx > 0) rect[count++] = { pw, 0, dx, hh };
+        if (dx < 0) rect[count++] = { 0, 0, -dx, hh };
+        for (int i = 0; i < count; ++i) {
+            XRectangle clip(rect[i]);
+            gfx.setClipRectangles(&clip, 1);
+            gfx.clearArea(clip.x, clip.y, clip.width, clip.height);
+            window()->paint(gfx, YRect(clip));
+        }
+        gfx.resetClip();
+        window()->setBackgroundPixmap(fPixmap);
+        window()->clearArea(0, 0, ww, hh);
+    }
+}
 
 // vim: set sw=4 ts=4 et:
