@@ -19,13 +19,11 @@
 YXApplication *xapp = nullptr;
 
 YDesktop *desktop = nullptr;
-YContext<YWindow> windowContext;
+YContext<YWindow> windowContext("windowContext", false);
 
-YCursor YXApplication::leftPointer;
-YCursor YXApplication::rightPointer;
-YCursor YXApplication::movePointer;
 bool YXApplication::synchronizeX11;
 bool YXApplication::alphaBlending;
+Window YXApplication::ignorable;
 
 Atom _XA_WM_CHANGE_STATE;
 Atom _XA_WM_CLASS;
@@ -68,6 +66,7 @@ Atom _XA_TARGETS;
 Atom _XA_XEMBED;
 Atom _XA_XEMBED_INFO;
 Atom _XA_UTF8_STRING;
+Atom _XA_COMPOUND_TEXT;
 
 Atom _XA_WIN_ICONS;
 Atom _XA_WIN_LAYER;
@@ -427,6 +426,7 @@ YAtomName YXApplication::atom_info[] = {
     { &_XA_XEMBED_INFO                      , "_XEMBED_INFO" },
     { &_XA_TARGETS                          , "TARGETS" },
     { &_XA_UTF8_STRING                      , "UTF8_STRING" },
+    { &_XA_COMPOUND_TEXT                    , "COMPOUND_TEXT" },
 
     { &XA_XdndActionAsk                     , "XdndActionAsk" },
     { &XA_XdndActionCopy                    , "XdndActionCopy" },
@@ -488,13 +488,6 @@ const char* YXApplication::atomName(Atom atom) {
     return buf;
 }
 
-void YXApplication::initPointers() {
-    osmart<YCursorLoader> l(YCursor::newLoader());
-    leftPointer  = l->load("left.xpm",  XC_left_ptr);
-    rightPointer = l->load("right.xpm", XC_right_ptr);
-    movePointer  = l->load("move.xpm",  XC_fleur);
-}
-
 void YXApplication::initModifiers() {
     XModifierKeymap *xmk = XGetModifierMapping(xapp->display());
     AltMask = MetaMask = WinMask = SuperMask = HyperMask =
@@ -531,8 +524,8 @@ void YXApplication::initModifiers() {
     if (MetaMask == AltMask)
         MetaMask = 0;
 
-    MSG(("alt:%d meta:%d super:%d hyper:%d mode:%d num:%d scroll:%d",
-         AltMask, MetaMask, SuperMask, HyperMask, ModeSwitchMask,
+    MSG(("alt:%d meta:%d super:%d hyper:%d win:%d mode:%d num:%d scroll:%d",
+         AltMask, MetaMask, SuperMask, HyperMask, WinMask, ModeSwitchMask,
          NumLockMask, ScrollLockMask));
 
     // some hacks for "broken" modifier configurations
@@ -803,9 +796,21 @@ bool YXApplication::filterEvent(const XEvent &xev) {
         XMappingEvent xmapping = xev.xmapping;
         XRefreshKeyboardMapping(&xmapping);
 
+        while (0 < XPending(display())) {
+            XEvent event;
+            XNextEvent(display(), &event);
+            if (event.type == MappingNotify) {
+                XRefreshKeyboardMapping(&event.xmapping);
+            } else {
+                XPutBackEvent(display(), &event);
+                break;
+            }
+        }
+
         initModifiers();
 
         desktop->grabKeys();
+        desktop->kbLayout();
         return true;
     }
     return false;
@@ -1024,8 +1029,10 @@ YXApplication::parseArgs(int argc, char **argv, const char *displayName) {
 Display* YXApplication::openDisplay(const char* displayName) {
     if (nonempty(displayName))
         setenv("DISPLAY", displayName, True);
+    else
+        displayName = getenv("DISPLAY");
 
-    Display* display = XOpenDisplay(nullptr);
+    Display* display = XOpenDisplay(displayName);
     if (display == nullptr)
         die(1, _("Can't open display: %s. X must be running and $DISPLAY set."),
             displayName ? displayName : _("<none>"));
@@ -1078,7 +1085,6 @@ YXApplication::YXApplication(int *argc, char ***argv, const char *displayName):
 
     initAtoms();
     initModifiers();
-    initPointers();
 }
 
 void YExtension::init(Display* dis, QueryFunc ext, QueryFunc ver) {
@@ -1118,8 +1124,8 @@ void YXApplication::initExtensions(Display* dpy) {
 }
 
 YXApplication::~YXApplication() {
-    if (fColormap32 != CopyFromParent)
-        XFreeColormap(xapp->display(), fColormap32);
+    if (fColormap32)
+        XFreeColormap(display(), fColormap32);
 
     xfd.unregisterPoll();
     XCloseDisplay(display());
@@ -1217,17 +1223,23 @@ void YXApplication::handleWindowEvent(Window xwindow, XEvent &xev) {
         }
     } else {
         if (xev.type == MapRequest) {
-            // !!! java seems to do this ugliness
-            //YFrameWindow *f = getFrame(xev.xany.window);
-            TLOG(("APP BUG? mapRequest for window %lX sent to destroyed frame %lX!",
-                xev.xmaprequest.parent,
-                xev.xmaprequest.window));
-            desktop->handleEvent(xev);
+            if (xev.xmaprequest.window != ignorable) {
+                // !!! java seems to do this ugliness
+                //YFrameWindow *f = getFrame(xev.xany.window);
+                TLOG(("APP BUG? mapRequest for window %lX "
+                      "sent to destroyed frame %lX!",
+                    xev.xmaprequest.window,
+                    xev.xmaprequest.parent));
+                desktop->handleEvent(xev);
+            }
         } else if (xev.type == ConfigureRequest) {
-            TLOG(("APP BUG? configureRequest for window %lX sent to destroyed frame %lX!",
-                xev.xmaprequest.parent,
-                xev.xmaprequest.window));
-            desktop->handleEvent(xev);
+            if (xev.xconfigurerequest.window != ignorable) {
+                TLOG(("APP BUG? configureRequest for window %lX "
+                      "sent to destroyed frame %lX!",
+                    xev.xconfigurerequest.window,
+                    xev.xconfigurerequest.parent));
+                desktop->handleEvent(xev);
+            }
         }
         else if (xev.type == ClientMessage && desktop) {
             Atom mesg = xev.xclient.message_type;
@@ -1262,7 +1274,7 @@ int YXApplication::errorHandler(Display* display, XErrorEvent* xev) {
     if (rc == Success)
         return rc;
 
-    XDBG {
+    XDBG if (xev->resourceid != ignorable) {
         char message[80], req[80], number[80];
 
         snprintf(number, sizeof number, "%d", xev->request_code);
@@ -1303,6 +1315,27 @@ int YXApplication::errorHandler(Display* display, XErrorEvent* xev) {
     return rc;
 }
 
+void YXApplication::send(XClientMessageEvent& ev, Window win, long mask) const {
+    XSendEvent(display(), win, False, mask, reinterpret_cast<XEvent*>(&ev));
+}
+
+Window YXApplication::parent(Window child) const {
+    Window paren = None;
+    Window rootw = None;
+    Window* data = nullptr;
+    unsigned num = None;
+    if (XQueryTree(display(), child, &rootw, &paren, &data, &num) && data) {
+        XFree(data);
+    }
+    return paren;
+}
+
+bool YXApplication::children(Window win, Window** data, unsigned* num) const {
+    Window rootw;
+    Window paren;
+    return XQueryTree(display(), win, &rootw, &paren, data, num);
+}
+
 void YXPoll::notifyRead() {
     owner()->handleXEvents();
 }
@@ -1326,18 +1359,31 @@ YAtom::operator Atom() {
 YTextProperty::YTextProperty(const char* str) {
     encoding = XA_STRING;
     format = 8;
-    if (str) {
-        nitems = strlen(str);
-        value = new unsigned char[1 + nitems];
-        if (value) memcpy(value, str, 1 + nitems);
-    } else {
-        nitems = 0;
-        value = nullptr;
+    value = str ? (unsigned char *) strdup(str) : nullptr;
+    nitems = value ? int(strlen((char *)value)) : 0;
+}
+
+YTextProperty::YTextProperty(Window handle, Atom property) {
+    nitems = 0;
+    value = nullptr;
+    if (XGetTextProperty(xapp->display(), handle, this, property)) {
+        if (encoding == _XA_COMPOUND_TEXT) {
+            char** list = nullptr;
+            int count = 0;
+            if (XmbTextPropertyToTextList(xapp->display(), this, &list, &count)
+                == Success && 0 < count) {
+                char* copy = strdup(*list);
+                XFreeStringList(list);
+                XFree(value);
+                value = (unsigned char *) copy;
+                encoding = XA_STRING;
+            }
+        }
     }
 }
 
 YTextProperty::~YTextProperty() {
-    if (value) delete[] value;
+    if (value) XFree(value);
 }
 
 void YProperty::discard() {
@@ -1362,4 +1408,16 @@ const YProperty& YProperty::update() {
     return *this;
 }
 
+void YProperty::append(void const* data, int count) const {
+    unsigned char const* bytes = reinterpret_cast<unsigned char const*>(data);
+    XChangeProperty(xapp->display(), fWind, fProp, fKind, fBits,
+                    PropModeAppend, bytes, count);
+}
+
+void YProperty::replace(void const* data, int count) const {
+    unsigned char const* bytes = reinterpret_cast<unsigned char const*>(data);
+    XChangeProperty(xapp->display(), fWind, fProp, fKind, fBits,
+                    PropModeReplace, bytes, count);
+}
+
 // vim: set sw=4 ts=4 et:
